# Snafu: Snake Functions - Messaging Connector

import kombu
import threading
import time
import os
import configparser

gcb = None

def initinternal(function, configpath):
	connecturl = None
	if not configpath:
		configpath = "snafu.ini"
	if not function:
		function = "snafu"
	if os.path.isfile(configpath):
		config = configparser.ConfigParser()
		config.read(configpath)
		if function in config and "connector.messaging" in config[function]:
			connecturl = config[function]["connector.messaging"]

	if not connecturl:
		return

	print("(messaging:connecting)")
	connection = kombu.Connection(connecturl)
	connection.connect()
	print("(messaging:connected)")

	queue = connection.SimpleQueue("sslq")
	while True:
		try:
			message = queue.get(block=True, timeout=20)
		except:
			print("(messaging:pass)")
			pass
		else:
			print("(messaging:received {})".format(message.payload))
			message.ack()
			# FIXME: send back response as AMQP message
			response = gcb(function, event="{}")
	queue.close()

def init(cb, function=None, configpath=None):
	global gcb
	gcb = cb

	t = threading.Thread(target=initinternal, daemon=True, args=(function, configpath))
	t.start()



# Snafu: Snake Functions - In-Memory Stateless (Isolated) Executor

import os
import time
import importlib.machinery

def execute(func, funcargs, envvars, sourceinfos):
	loader = importlib.machinery.SourceFileLoader(os.path.basename(sourceinfos.source), sourceinfos.source)
	loader.exec_module(sourceinfos.module)

	for envvar in envvars:
		os.environ[envvar] = envvars[envvar]

	stime = time.time()
	try:
		res = func(*funcargs)
		success = True
	except Exception as e:
		res = e
		success = False
	dtime = (time.time() - stime) * 1000

	return dtime, success, res



# Snafu: Snake Functions - In-Memory Executor

import os
import time

def execute(func, funcargs, envvars, sourceinfos):
	for envvar in envvars:
		os.environ[envvar] = envvars[envvar]

	stime = time.time()
	try:
		res = func(*funcargs)
		success = True
	except Exception as e:
		res = e
		success = False
	dtime = (time.time() - stime) * 1000

	return dtime, success, res



# Snafu: Snake Functions - Lambda Authenticator

import hmac
import hashlib
import base64
import os.path
import configparser
import binascii

def hmacsha256(secret, msg):
	return hmac.new(secret, msg=msg, digestmod=hashlib.sha256).digest()

def authorise(flaskrequest):
	auth = flaskrequest.headers.get("Authorization")
	host = flaskrequest.headers.get("Host")
	amzdate = flaskrequest.headers.get("X-Amz-Date")
	if not auth or not host or not amzdate:
		return False

	path = flaskrequest.path
	method = flaskrequest.method

	auth = auth.split(" ")
	if len(auth) != 4:
		return False

	algo, cred, sh, sig = auth
	if algo != "AWS4-HMAC-SHA256":
		return False

	cred = cred.split("=")
	sh = sh.split("=")
	sig = sig.split("=")
	if len(cred) != 2 or len(sh) != 2 or len(sig) != 2:
		return False
	if cred[0] != "Credential" or sh[0] != "SignedHeaders" or sig[0] != "Signature":
		return False
	cred = cred[1][:-1]
	sh = sh[1][:-1]
	sig = sig[1]

	cred = cred.split("/")
	if len(cred) != 5:
		return False
	keyid, date, region, service, request = cred

	c = configparser.ConfigParser()
	try:
		accdb = os.path.expanduser("~/.snafu-accounts")
		if os.getenv("HOME") == "/":
			accdb = "/root/.snafu-accounts"
		c.read(accdb)
	except:
		return False
	sections = c.sections()
	secretkey = None
	for section in sections:
		ckeyid = c.get(section, "access_key_id")
		if ckeyid == keyid:
			secretkey = c.get(section, "secret_access_key")
			break
	if not secretkey:
		return False

	enc = "utf-8"

	b_secretkey = bytes("AWS4" + secretkey, enc)
	b_date = bytes(date, enc)
	b_region = bytes(region, enc)
	b_service = bytes(service, enc)
	b_request = bytes(request, enc)

	ch = ""
	ch += "host:{}".format(host) + "\n"
	ch += "x-amz-date:{}".format(amzdate) + "\n"

	# FIXME: check existing sh for required fields
	#sh = "host" + ";" + "x-amz-date"

	payload = flaskrequest.data
	hp = hashlib.sha256(payload).hexdigest()

	cr = ""
	cr += method + "\n"
	cr += path + "\n"
	cr += "" + "\n"
	cr += ch + "\n"
	cr += sh + "\n"
	cr += hp

	hcr = base64.b16encode(hashlib.sha256(bytes(cr, "utf-8")).digest()).decode("utf-8").lower()

	sts = ""
	sts += algo + "\n"
	sts += amzdate + "\n"
	sts += "{}/{}/{}/{}".format(date, region, service, request) + "\n"
	sts += hcr

	b_string = bytes(sts, enc)

	dig_dk = hmacsha256(b_secretkey, b_date)
	dig_drk = hmacsha256(dig_dk, b_region)
	dig_drsk = hmacsha256(dig_drk, b_service)
	dig_sk = hmacsha256(dig_drsk, b_request)

	dig_sig = hmacsha256(dig_sk, b_string)

	signature = binascii.hexlify(dig_sig).decode("utf-8")

	#print("SIGNATURE", signature)
	#print("MATCH    ", sig)

	#return True
	return keyid



# Snafu: Snake Functions - Proxy Executor

import requests
import os
import configparser

def executecontrol(flaskrequest, tenant):
	endpoint = None

	if tenant:
		c = configparser.ConfigParser()
		try:
			accdb = os.path.expanduser("~/.snafu-accounts")
			if os.getenv("HOME") == "/":
				accdb = "/root/.snafu-accounts"
			c.read(accdb)
		except:
			return
		sections = c.sections()
		for section in sections:
			ckeyid = c.get(section, "access_key_id")
			if ckeyid == tenant:
				endpoint = c.get(section, "endpoint")
				break
	else:
		config = configparser.ConfigParser()
		config.read("snafu.ini")
		if "snafu" in config and "executor.proxy" in config["snafu"]:
			endpoint= config["snafu"]["executor.proxy"]
	if not endpoint:
		return

	headers = {}
	headers["X-Amz-Date"] = flaskrequest.headers.get("X-Amz-Date")

	data = flaskrequest.data.decode("utf-8")
	#method=r.method -> requests.post

	reply = requests.post(endpoint + flaskrequest.path, data=data, headers=headers)

	if reply.status_code == 200:
		return reply.content.decode("utf-8")
	else:
		return



# Snafu: Snake Functions - Python2 Executor

import json
import subprocess
import pickle
import base64
import tempfile

def strbool(x):
	return True if x == "True" else False

def execute(func, funcargs, envvars, sourceinfos):
	for i, funcarg in enumerate(funcargs):
		if "__class__" in dir(funcarg) and funcarg.__class__.__name__ == "SnafuContext":
			funcargs[i] = "pickle:" + base64.b64encode(pickle.dumps(funcarg, protocol=2)).decode("utf-8")

	funcargs = json.dumps(funcargs)
	envvars = json.dumps(envvars)

	if len(funcargs) > 256:
		tf = tempfile.NamedTemporaryFile()
		tf.write(bytes(funcargs, "utf-8"))
		funcargs = "tempfile:" + tf.name

	p = subprocess.run("python2 executors/python2-exec.py {} {} '{}' '{}'".format(sourceinfos.source, func.__name__, funcargs, envvars), stdout=subprocess.PIPE, shell=True)

	try:
		dtime, success, *res = p.stdout.decode("utf-8").strip().split(" ")
	except:
		dtime = 0.0
		success = False
		res = []
	dtime = float(dtime)
	success = strbool(success)
	res = " ".join(res)
	#print("PY2", res)
	return dtime, success, res



# Snafu: Snake Functions - Python2 Executor (Exec Module)

import sys
import os
import time
import json
import base64
import pickle

class Context:
	def __init__(self):
		self.SnafuContext = self

	def __new__(self, x):
		return x

def execute(filename, func, funcargs, envvars):
	if funcargs.startswith("tempfile:"):
		funcargs = open(funcargs[9:]).read()

	funcargs = json.loads(funcargs)
	envvars = json.loads(envvars)

	for i, funcarg in enumerate(funcargs):
		if type(funcarg) == str and funcarg.startswith("pickle:"):
			sys.modules["lib"] = Context()
			sys.modules["lib.snafu"] = Context()
			funcarg = pickle.loads(base64.b64decode(funcarg.split(":")[1]))

	sys.path.append(".")
	os.chdir(os.path.dirname(filename))
	mod = __import__(os.path.basename(filename[:-3]))
	func = getattr(mod, func)

	for envvar in envvars:
		os.environ[envvar] = envvars[envvar]

	stime = time.time()
	try:
		res = func(*funcargs)
		success = True
	except Exception as e:
		res = e
		success = False
	dtime = (time.time() - stime) * 1000

	#return dtime, success, res
	return "{} {} {}".format(dtime, success, "{}".format(res).replace("'", "\""))

print(execute(sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4]))



# Snafu: Snake Functions - Account Management

import sys
import os.path
import argparse
import configparser

class SnafuAccounts:
	def __init__(self):
		self.accounts = {}

	def addaccount(self, key, secret, endpoint):
		accdb = os.path.expanduser("~/.snafu-accounts")
		if os.getenv("HOME") == "/":
			accdb = "/root/.snafu-accounts"
		c = configparser.ConfigParser()
		try:
			c.read(accdb)
		except:
			pass
		acc = "account{}".format(len(c.sections()))
		c.add_section(acc)
		c.set(acc, "access_key_id", key)
		c.set(acc, "secret_access_key", secret)
		if endpoint:
			c.set(acc, "endpoint", endpoint)
		f = open(accdb, "w")
		c.write(f)

	def run(self):
		parser = argparse.ArgumentParser(description="Snake Functions as a Service - Account Management")
		parser.add_argument("-a", "--add", help="add an account", action="store_true")
		parser.add_argument("-k", "--key-id", help="access key id")
		parser.add_argument("-s", "--secret-access-key", help="secret access key")
		parser.add_argument("-e", "--endpoint", help="tenant-specific static endpoint")
		args = parser.parse_args()

		if args.add:
			if not args.key_id or not args.secret_access_key:
				print("Error: Must specify both --key-id and --secret-access-key for --add.", file=sys.stderr)
				return

			self.addaccount(args.key_id, args.secret_access_key, args.endpoint)
		else:
			print("Error: Must specify --add.", file=sys.stderr)
			return

class SnafuAccountsRunner:
	def __init__(self):
		sa = SnafuAccounts()
		sa.run()



# Snafu: Snake Functions - Control Plane Module

import flask
import json
import argparse
import os
import base64
import subprocess
import glob
import snafulib.snafu
import logging
import importlib
import threading
import time

"""
AWS CLI support
Legend: * = implemented, / = noop, # = not necessary
* [s3] "download-function"
/ [sts] get-caller-identity
/ add-permission
create-alias
create-event-source-mapping
* create-function
delete-alias
delete-event-source-mapping
delete-function
get-account-settings
get-alias
get-event-source-mapping
* get-function
* get-function-configuration
get-policy
# help
* invoke
invoke-async
list-aliases
list-event-source-mappings
* list-functions
list-versions-by-function
publish-version
remove-permission
update-alias
update-event-source-mapping
* update-function-code
* update-function-configuration

GCloud CLI support
Legend: * = implemented, / = noop, # = not necessary
(*) call
delete
deploy
describe
* list

WSK CLI support
Legend: * = implemented, / = noop, # = not necessary
/ namespace list
(*) list
(*) action list
(*) action get
(*) action invoke
"""

class ControlUtil:
	def createconfig(function, sourceinfos):
		sourcename, functionname = function.split(".")

		fmap = {}
		env = {}
		envvars = {}
		if envvars:
			env["Variables"] = envvars
		vpc = {}
		#vpc["SecurityGroupIds"] = []
		#vpc["SubnetIds"] = []
		fmap["FunctionName"] = function
		fmap["Version"] = "$LATEST"
		#fmap["MemorySize"] = 128
		fmap["MemorySize"] = -1
		fmap["Handler"] = "{}.{}".format(sourcename, functionname)
		fmap["LastModified"] = "2017-01-01T00:00:00.000+0000"
		fmap["CodeSha256"] = sourceinfos.checksum
		fmap["CodeSize"] = sourceinfos.size
		fmap["Timeout"] = 3
		#fmap["Runtime"] = "python2.7"
		fmap["Runtime"] = "python3"
		fmap["Description"] = "<function without configuration>"
		#fmap["Role"] = "arn:...:XXX"
		#fmap["FunctionArn"] = "arn:...:XXX"
		if env:
			fmap["Environment"] = env
		if vpc:
			fmap["VpcConfig"] = vpc
		return fmap

	def functionconfiguration(function):
		func, config, sourceinfos = SnafuControl.snafu.functions[function]
		if config:
			fmap = config
		else:
			fmap = ControlUtil.createconfig(function, sourceinfos)
		return fmap

	def notauthorised():
		err = json.dumps({"errorMessage": "NotAuthenticated"})
		return err, 501

	def reaperthread(debug=False):
		if debug:
			print("REAPER init")
		f = open("/proc/net/tcp")
		while True:
			time.sleep(1)
			try:
				ControlUtil.reaperthread_internal(True, f)
			except Exception as e:
				if debug:
					print("REAPER crash {}".format(e))

	def deployerthread(snafu, debug=False):
		if debug:
			print("DEPLOYER init")
		import pyinotify
		wm = pyinotify.WatchManager()
		mask = pyinotify.IN_CREATE

		class EventHandler(pyinotify.ProcessEvent):
			def process_IN_CREATE(self, event):
				print("Hot deployment of function {}".format(event.pathname))
				snafu.activate([event.pathname], "any")

		handler = EventHandler()
		notifier = pyinotify.Notifier(wm, handler)
		wdd = wm.add_watch("functions-local", mask, rec=True)
		notifier.loop()

	def reaperthread_internal(debug, f):
		if debug:
			print("REAPER reap")
		fds = glob.glob("/proc/{}/fd/*".format(os.getpid()))
		sockets = {}
		for fd in fds:
			try:
				rp = os.path.realpath(fd)
				if os.path.islink(fd) and not os.path.exists(rp):
					socket = int(os.path.basename(fd))
					if debug:
						print("+", socket, rp)
					rp = os.path.basename(rp).replace("socket:[", "").replace("]", "")
					sockets[rp] = socket
			except:
				if debug:
					print("-", fd)
				pass
		f.seek(0)
		tcp = f.readlines()
		for line in tcp:
			line = [x for x in line.split(" ") if x]
			st = line[3]
			inode = line[9]
			if inode in sockets:
				if debug:
					print("?", inode, st)
				if st == "08":
					if debug:
						print("! close", sockets[inode])
					else:
						print("Reap dangling connection {}".format(sockets[inode]))
					try:
						os.close(sockets[inode])
					except:
						pass

	def reaper():
		t = threading.Thread(target=ControlUtil.reaperthread, args=(False,))
		t.setDaemon(True)
		t.start()

	def deployer(snafu):
		t = threading.Thread(target=ControlUtil.deployerthread, args=(snafu, False))
		t.setDaemon(True)
		t.start()

class SnafuControl:
	app = flask.Flask("snafu-control")
	port = None
	snafu = None
	authenticatormods = []

	def __init__(self):
		pass

	@app.route("/<user>-<namespace>-<number>", methods=["HEAD", "PUT"])
	def s3(user, namespace, number):
		#if flask.request.method == "HEAD":
		#	return "{}", 404
		return "{}"

	@app.route("/<bucket>/<dir>/<objectname>.meta.json")
	def s3meta(bucket, dir, objectname):
		return "{\"python_ver\": \"3.5\"}"

	@app.route("/", methods=["POST"])
	def sts_iam():
		if "Action" in flask.request.form:
			action = flask.request.form["Action"]
			if action == "GetCallerIdentity":
				return "<GetCallerIdentityResponse xmlns='https://sts.amazonaws.com/doc/2011-06-15/'><GetCallerIdentityResult><Account>000000000000</Account></GetCallerIdentityResult></GetCallerIdentityResponse>"
			elif action == "CreateRole":
				return "<CreateRoleResponse xmlns='https://iam.amazonaws.com/doc/2010-05-08/'><CreateRoleResult><Role></Role></CreateRoleResult></CreateRoleResponse>"
			elif action == "PutRolePolicy":
				return "<PutRolePolicyResponse xmlns='https://iam.amazonaws.com/doc/2010-05-08/'></PutRolePolicyResponse>"
			else:
				err = json.dumps({"errorMessage": "UndefinedAction"})
				return err, 501

	@app.route("/api/v1/namespaces")
	def listnamespaceopenwhisk():
		d = ["snafu"]
		return json.dumps(d)

	@app.route("/api/v1/namespaces/<namespace>/")
	def listfunctionsopenwhisk(namespace):
		overridenamespace = "snafu"
		f = []
		for function in SnafuControl.snafu.functions:
			ann = [{"key": "exec", "value": "python:2"}]
			fmap = {"namespace": overridenamespace, "name": function, "annotations": ann}
			f.append(fmap)
		d = {"actions": f}
		return json.dumps(d)

	@app.route("/api/v1/namespaces/<namespace>/actions")
	# ?limit=30&skip=0
	def listfunctionsactionsopenwhisk(namespace):
		overridenamespace = "snafu"
		f = []
		for function in SnafuControl.snafu.functions:
			fmap = {"namespace": overridenamespace, "name": function}
			f.append(fmap)
		return json.dumps(f)

	@app.route("/api/v1/namespaces/<namespace>/actions/<function>")
	def getfunctionopenwhisk(namespace, function):
		overridenamespace = "snafu"
		d = {"namespace": overridenamespace, "name": function, "version": "0.0.1", "publish": False}
		return json.dumps(d)

	@app.route("/api/v1/namespaces/<namespace>/actions/<function>", methods=["POST"])
	# ?blocking=true&result=true
	def invokeopenwhisk(namespace, function):
		data = flask.request.data.decode("utf-8")
		dataargs = json.loads(data)
		#print("F", function, dataargs)
		response = SnafuControl.snafu.execute(function, **dataargs)
		#print("R", response)
		if type(response) == dict:
			response = json.dumps(response)
		return response

	@app.route("/v1beta2/projects/<project>/locations/<location>/functions")
	# alt=json&pageSize=100
	def listfunctionsgoogle(project, location):
		f = []
		for function in SnafuControl.snafu.functions:
			fmap = {"name": function, "status": "ready", "httpsTrigger": {"url": "..."}}
			f.append(fmap)

		functions = {"functions" : f}
		return json.dumps(functions)

	@app.route("/v1beta2/projects/<project>/locations/<location>/functions/<function>", methods=["POST"])
	def invokegoogle(project, location, function):
		function = function.split(":")[0]
		response = SnafuControl.snafu.execute(function)
		return json.dumps(response)

	@app.route("/function-download/<function>.zip")
	def functiondownload(function):
		func, config, sourceinfos = SnafuControl.snafu.functions[function]
		try:
			content = open(sourceinfos.source).read()
		except:
			err = json.dumps({"errorMessage": "NoZipFilePresent"})
			return err, 501
		return flask.Response(content, mimetype="application/octet-stream")

	@app.route("/2015-03-31/functions", methods=["POST"])
	def createfunction():
		auth = SnafuControl.authorise()
		if not auth:
			return ControlUtil.notauthorised()

		data = flask.request.data.decode("utf-8")
		config = json.loads(data)
		functionname = config["FunctionName"]

		if auth[1]:
			snafulib.snafu.SnafuImport.functiondir = "functions-tenants/{}/{}".format(auth[1], functionname)
		else:
			snafulib.snafu.SnafuImport.functiondir = "functions-local/{}".format(functionname)

		try:
			#os.makedirs(snafulib.snafu.SnafuImport.functiondir)
			snafulib.snafu.SnafuImport.prepare()
		except:
			err = json.dumps({"errorMessage": "ResourceConflictException"})
			return err, 501

		codezip = "{}.zip".format(snafulib.snafu.SnafuImport.functiondir)
		f = open(codezip, "wb")
		f.write(base64.b64decode(config["Code"]["ZipFile"]))
		f.close()

		del config["Code"]
		### TODO: extend configuration

		#snafulib.snafu.SnafuImport.prepare()
		snafulib.snafu.SnafuImport.importfunction(functionname, codezip, config, convert=True)

		return json.dumps(config)

	@app.route("/2015-03-31/functions/")
	def listfunctions():
		auth = SnafuControl.authorise()
		if not auth:
			return ControlUtil.notauthorised()

		f = []
		for function in SnafuControl.snafu.functions:
			fmap = ControlUtil.functionconfiguration(function)
			f.append(fmap)

		functions = {"Functions" : f}
		return json.dumps(functions)

	@app.route("/2015-03-31/functions/<function>")
	def getfunction(function):
		auth = SnafuControl.authorise()
		if not auth:
			return ControlUtil.notauthorised()

		if function in SnafuControl.snafu.functions:
			funccode = {}
			funccode["RepositoryType"] = "snafu"
			funccode["Location"] = "http://localhost:{}/function-download/{}.zip".format(SnafuControl.port, function)
			func = {}
			func["Code"] = funccode
			func["Configuration"] = ControlUtil.functionconfiguration(function)
			return json.dumps(func)
		else:
			err = json.dumps({"errorMessage": "ResourceNotFoundException"})
			return err, 501

	@app.route("/2015-03-31/functions/<function>/code", methods=["PUT"])
	def putfunctioncode(function):
		auth = SnafuControl.authorise()
		if not auth:
			return ControlUtil.notauthorised()

		if function in SnafuControl.snafu.functions:
			data = flask.request.data.decode("utf-8")
			update = json.loads(data)

			if auth[1]:
				snafulib.snafu.SnafuImport.functiondir = "functions-tenants/{}/{}".format(auth[1], function)
			else:
				snafulib.snafu.SnafuImport.functiondir = "functions-local/{}".format(function)

			codezip = "{}.zip".format(snafulib.snafu.SnafuImport.functiondir)
			f = open(codezip, "wb")
			f.write(base64.b64decode(update["ZipFile"]))
			f.close()

			snafulib.snafu.SnafuImport.importfunction(function, codezip, None)
			func, config, sourceinfos = SnafuControl.snafu.functions[function]
			#fmap = ControlUtil.functionconfiguration(function)
			# FIXME: update config, sourceinfos.checksum, sourceinfos.size
			return json.dumps(config)
		else:
			err = json.dumps({"errorMessage": "ResourceNotFoundException"})
			return err, 501

	@app.route("/2015-03-31/functions/<function>/configuration", methods=["GET", "PUT"])
	def getputfunctionconfiguration(function):
		auth = SnafuControl.authorise()
		if not auth:
			return ControlUtil.notauthorised()

		if function in SnafuControl.snafu.functions:
			if flask.request.method == "GET":
				return json.dumps(ControlUtil.functionconfiguration(function))
			elif flask.request.method == "PUT":
				data = flask.request.data.decode("utf-8")
				update = json.loads(data)
				fmap = ControlUtil.functionconfiguration(function)
				for key in update:
					fmap[key] = update[key]
				snafulib.snafu.SnafuImport.importfunction(function, None, fmap)
				return json.dumps(fmap)
		else:
			err = json.dumps({"errorMessage": "ResourceNotFoundException"})
			return err, 501

	@app.route("/2015-03-31/functions/<function>/policy", methods=["POST"])
	def addpermission(function):
		auth = SnafuControl.authorise()
		if not auth:
			return ControlUtil.notauthorised()

		data = flask.request.data.decode("utf-8")
		"""
		{"Action": "lambda:InvokeFunction", "StatementId": "Y_reverse", "Principal": "X"}
		->
		{"Statement": "{\"Sid\":\"Y_reverse\",\"Resource\":\"arn:aws:lambda:us-west-1:*:function:Y\",\"Effect\":\"Allow\",\"Principal\":{\"AWS\":\"arn:aws:iam::*:role/lambda_basic_execution\"},\"Action\":[\"lambda:InvokeFunction\"]}"}
		"""
		return "{}"

	@app.route("/2015-03-31/functions/<function>/invocations", methods=["POST"])
	def invoke(function):
		auth = SnafuControl.authorise()
		if not auth:
			return ControlUtil.notauthorised()

		if SnafuControl.snafu.executormods[0].__name__ == "executors.docker":
			response = SnafuControl.snafu.executormods[0].executecontrol(flask.request, auth[1])
		else:
			dataargs = {}
			data = flask.request.data.decode("utf-8")
			if data:
				dataargs["event"] = json.loads(data)

			response = SnafuControl.snafu.execute(function, **dataargs)

		if not response:
			#flask.abort(500)
			err = json.dumps({"errorMessage": "ServiceException"})
			return err, 501
		if type(response) == dict:
			response = json.dumps(response)
		return response

	@app.errorhandler(404)
	def notfound(error):
		return "Invalid request.", 404

	def authorise():
		if len(SnafuControl.authenticatormods) == 0:
			return True, None

		for authenticatormod in SnafuControl.authenticatormods:
			auth = authenticatormod.authorise(flask.request)
			if auth:
				if not SnafuControl.snafu.quiet:
					print("Authorised: {}".format(auth))
				return True, auth

		return False

	def setupauthenticators(self, authenticators, quiet=False):
		if "none" in authenticators:
			return

		if not quiet:
			for authenticator in authenticators:
				print("+ authenticator:", authenticator)

		SnafuControl.authenticatormods = []
		for authenticator in authenticators:
			mod = importlib.import_module("authenticators." + authenticator)
			SnafuControl.authenticatormods.append(mod)

	def run(self):
		parser = argparse.ArgumentParser(description="Snake Functions as a Service - Control Plane")
		snafulib.snafu.SnafuRunner.add_common_arguments(parser)
		#parser.add_argument("-i", "--isolation", help="isolate stateful function instances", action="store_true")
		parser.add_argument("-a", "--authenticator", help="function authenticators; 'none' by default", choices=["aws", "none"], default=["none"], nargs="+")
		parser.add_argument("-p", "--port", help="HTTP port number", type=int, default=10000)
		parser.add_argument("-r", "--reaper", help="closed connection reaper", action="store_true")
		parser.add_argument("-d", "--deployer", help="automated hot deployment", action="store_true")

		for action in parser._actions:
			if action.dest == "executor":
				action.choices.append("docker")
				action.choices.append("proxy")
				action.choices.append("openshift")

		args = parser.parse_args()

		ignore = False
		if not args.file:
			args.file.append("functions")
			args.file.append("functions-local")
			args.file.append("functions-tenants")
			ignore = True

		SnafuControl.snafu = snafulib.snafu.Snafu(args.quiet)
		SnafuControl.port = args.port
		# FIXME: should now provide the choice between lambda, gfunctions and openwhisk depending on modular activation
		self.snafu.activate(args.file, "lambda", ignore=ignore)
		self.snafu.setuploggers(args.logger)
		self.snafu.setupexecutors([args.executor])
		self.setupauthenticators(args.authenticator, args.quiet)

		if args.quiet:
			log = logging.getLogger("werkzeug")
			log.setLevel(logging.ERROR)

		if args.reaper:
			print("+ reaper")
			ControlUtil.reaper()

		if args.deployer:
			print("+ deployer")
			ControlUtil.deployer(SnafuControl.snafu)

		self.app.run(host="0.0.0.0", port=args.port, threaded=True)

class SnafuControlRunner:
	def __init__(self):
		try:
			import setproctitle
			setproctitle.setproctitle("snafu-control")
		except:
			pass

		sc = SnafuControl()
		sc.run()



# Snafu: Snake Functions - Docker Executor

import requests
import os
import configparser
import time
import random

container = "jszhaw/snafu"

endpoints = {}
multi = 3
authorised = True

def launch(tenant, portnum):
	authmount = ""
	if authorised:
		accdb = os.path.expanduser("~/.snafu-accounts")
		accdir = os.path.expanduser("~/.snafu-accounts-tenants/{}".format(tenant))
		if os.getenv("HOME") == "/":
			accdb = "/root/.snafu-accounts"
			accdir = "/root/.snafu-accounts-tenants/{}".format(tenant)
		c = configparser.ConfigParser()
		try:
			c.read(accdb)
		except:
			pass
		else:
			sections = c.sections()
			secretkey = None
			for section in sections:
				ckeyid = c.get(section, "access_key_id")
				if ckeyid == tenant:
					secretkey = c.get(section, "secret_access_key")
					break

			if secretkey:
				authmount = "-v {}:/root/.aws".format(accdir)
				os.makedirs(accdir, exist_ok=True)

				c = configparser.ConfigParser()
				acc = "default"
				c.add_section(acc)
				c.set(acc, "aws_access_key_id", tenant)
				c.set(acc, "aws_secret_access_key", secretkey)
				f = open("{}/credentials".format(accdir), "w")
				c.write(f)
				f.close()

				c = configparser.ConfigParser()
				acc = "default"
				c.add_section(acc)
				c.set(acc, "region", "invalid")
				f = open("{}/config".format(accdir), "w")
				c.write(f)
				f.close()
	dockercmd = "docker run -d -p 127.0.0.1:{}:10000 -v /opt/functions-tenants/{}:/opt/functions-local {} {}".format(portnum, tenant, authmount, container)
	os.system(dockercmd)

def executecontrol(flaskrequest, tenant):
	if not tenant in endpoints:
		if multi:
			endpoints[tenant] = []
			for i in range(multi):
				entries = sum([len(x) if type(x) == list else 1 for x in endpoints.values()])
				portnum = 10000 + entries + 1
				endpoints[tenant].append("http://localhost:{}".format(portnum))
				launch(tenant, portnum)
		else:
			entries = len(endpoints)
			portnum = 10000 + entries + 1
			endpoints[tenant] = "http://localhost:{}".format(portnum)
			launch(tenant, portnum)
		time.sleep(2)

	endpoint = endpoints[tenant]
	if multi:
		endpoint = random.choice(endpoint)
		#print("CHOICE", endpoint)

	headers = {}
	headers["X-Amz-Date"] = flaskrequest.headers.get("X-Amz-Date")

	data = flaskrequest.data.decode("utf-8")
	#method=r.method -> requests.post

	reply = requests.post(endpoint + flaskrequest.path, data=data, headers=headers)

	if reply.status_code == 200:
		return reply.content.decode("utf-8")
	else:
		return



# Snafu: Snake Functions - JavaScript Executor

import subprocess
import time

def execute(func, funcargs, envvars, sourceinfos):
	sourcemodule = sourceinfos.source[:-3]

	cmd = "nodejs -e 'require(\"./%s\").%s({\"body\": {\"message\": \"%s\"}}, {\"status\": function(x){return {\"send\": function(x){console.log(\"RET:\" + x)}}}})'" % (sourcemodule, func, funcargs[0])

	stime = time.time()
	p = subprocess.run(cmd, stdout=subprocess.PIPE, shell=True)
	out = p.stdout.decode("utf-8")
	dtime = time.time() - stime

	success = False
	res = []
	for line in out.split("\n"):
		if line.startswith("RET:"):
			success = True
			res = line[4:]

	return dtime, success, res



# Snafu: Snake Functions - CSV Logger

import os
import configparser

def log(source, function, duration, success):
	logurl = "snafu.csv"
	if os.path.isfile("snafu.ini"):
		config = configparser.ConfigParser()
		config.read("snafu.ini")
		if "snafu" in config and "logger.csv" in config["snafu"]:
			logurl = config["snafu"]["logger.csv"]

	# FIXME: this part must become mt-safe through a protected section
	f = open(logurl, "a")
	print("{},{},{:1.3f},{}".format(source, function, duration, success), file=f)
	f.close()



# Snafu: Snake Functions - Import Utility Module

import argparse
import subprocess
import json
import os
import sys
import tempfile
import glob
import uuid

import snafulib.snafu

class SnafuImportUtility:
	def __init__(self):
		self.td = tempfile.TemporaryDirectory()

	def start_import(self):
		parser = argparse.ArgumentParser(description="Snake Functions as a Service - Import Utility")
		parser.add_argument("-s", "--source", help="import source", choices=["lambda", "gfunctions", "openwhisk"], default=None)
		parser.add_argument("-t", "--target", help="import target", choices=["snafu", "funktion", "fission", "kubeless"], default="snafu")
		parser.add_argument("-c", "--convert", help="convert functions from Python 2 to Python 3 for native execution")
		args = parser.parse_args()

		snafulib.snafu.SnafuImport.prepare()

		if args.source is None:
			print("Must specify a source with --source.", file=sys.stderr)
		elif args.source == "lambda":
			self.import_lambda(args.target, args.convert)
		elif args.source == "gfunctions":
			self.import_gfunctions(args.target)
		elif args.source == "openwhisk":
			self.import_openwhisk(args.target)

	def export_snafu(self, filename, code, name, env):
		#codefile, configfile, oldcodefile = snafulib.snafu.SnafuImport.importfunction(actioninfo["name"], None, x, False)
		codefile = os.path.join(snafulib.snafu.SnafuImport.functiondir, filename)
		f = open(codefile, "w")
		f.write(code)
		print("+ code: {}".format(codefile))

	def export_snafu_zip(self, filename, code, name, env, config=None, convert=False):
		codefile, configfile, oldcodefile = snafulib.snafu.SnafuImport.importfunction(name, filename, config, convert)
		print("+ code: {}".format(codefile))
		print("+ configuration: {}".format(configfile))
		if oldcodefile:
			print("+ converted; old code: {}".format(oldcodefile))

	def archivedfile(self, filename):
		td = os.path.join(self.td.name, str(uuid.uuid4()))
		os.makedirs(td, exist_ok=True)
		cwd = os.getcwd()
		subprocess.run("cd {} && unzip -o -q {}".format(td, os.path.join(cwd, filename)), shell=True)
		codefiles = glob.glob(os.path.join(td, "*.py"))
		if len(codefiles) == 0:
			codefiles = glob.glob(os.path.join(td, "*.java"))
		if len(codefiles) == 0:
			codefiles = glob.glob(os.path.join(td, "*.class"))
		if len(codefiles) == 0:
			codefiles = glob.glob(os.path.join(td, "*.js"))
		return codefiles

	def export_funktion(self, filename, code, name, env):
		if not code:
			codefiles = self.archivedfile(filename)
			if len(codefiles) == 1:
				filename = codefiles[0]
				code = open(filename).read()
			else:
				print("- skip archive: ambiguous contents ({})".format(codefiles))
				return
		codeline = code.replace("\n", "\\n")
		subprocess.run("funktion create fn -n '{}' -s '{}'".format(name, codeline), shell=True)
		print("+ code: {} (in funktion)".format(os.path.basename(filename)))

	def export_fission(self, filename, code, name, env):
		if not code:
			codefiles = self.archivedfile(filename)
			if len(codefiles) == 1:
				filename = codefiles[0]
				code = open(filename).read()
			else:
				print("- skip archive: ambiguous contents ({})".format(codefiles))
				return
		f = tempfile.NamedTemporaryFile()
		f.write(bytes(code, "utf-8"))
		f.flush()
		codefile = f.name
		subprocess.run("fission function create --name '{}' --env {} --code '{}'".format(name, env, codefile), shell=True)
		print("+ code: {} (in fission)".format(os.path.basename(filename)))

	def export_kubeless(self, filename, code, name, env):
		if not code:
			codefiles = self.archivedfile(filename)
			if len(codefiles) == 1:
				filename = codefiles[0]
				code = open(filename).read()
			else:
				print("- skip archive: ambiguous contents ({})".format(codefiles))
				return
		f = tempfile.NamedTemporaryFile()
		f.write(bytes(code, "utf-8"))
		f.flush()
		codefile = f.name
		mangledname = name.replace(" ", "-").lower()
		handler = os.path.basename(codefile) + ".main"
		subprocess.run("kubeless function deploy '{}' --runtime {} --handler '{}' --from-file '{}' --trigger-http".format(mangledname, env, handler, codefile), shell=True)
		print("+ code: {} (in kubeless)".format(os.path.basename(filename)))

	def import_openwhisk(self, target):
		proc = subprocess.run("wsk list", stdout=subprocess.PIPE, shell=True)
		out = proc.stdout.decode("utf-8")
		inactions = False
		for line in out.split("\n"):
			line = line.split(" " * 2)
			if line[0] == "actions":
				inactions = True
			elif line[0] == "triggers":
				inactions = False
			elif inactions:
				function = line[0]
				print("import", function)
				proc = subprocess.run("wsk action get '{}'".format(function), stdout=subprocess.PIPE, shell=True)
				out = proc.stdout.decode("utf-8")
				out = out[out.find("\n"):]
				#print("CODE", out)
				actioninfo = json.loads(out)
				code = actioninfo["exec"]
				filename = None
				env = None
				if code["kind"] == "nodejs" or code["kind"] == "nodejs:6":
					filename = actioninfo["name"] + ".js"
					env = "nodejs"
				elif code["kind"] == "python" or code["kind"] == "python:2":
					filename = actioninfo["name"] + ".py"
					env = "python"
				else:
					print("- skip ({})".format(code["kind"]))

				if filename:
					if target == "snafu":
						self.export_snafu(filename, code["code"], actioninfo["name"], env)
					elif target == "funktion":
						self.export_funktion(filename, code["code"], actioninfo["name"], env)
					elif target == "fission":
						self.export_fission(filename, code["code"], actioninfo["name"], env)
					elif target == "kubeless":
						self.export_kubeless(filename, code["code"], actioninfo["name"], env)

	def import_gfunctions(self, target):
		proc = subprocess.run("gcloud beta functions list", stdout=subprocess.PIPE, shell=True)
		out = proc.stdout.decode("utf-8")
		functiondir = snafulib.snafu.SnafuImport.functiondir
		for line in out.split("\n"):
			lineparts = line.split(" ")
			if len(lineparts) > 1 and lineparts[0] != "NAME":
				funcname = lineparts[0]
				print("import", funcname)
				snafulib.snafu.SnafuImport.functiondir = os.path.join(functiondir, funcname)
				proc = subprocess.run("gcloud beta functions describe {}".format(funcname), stdout=subprocess.PIPE, shell=True)
				out = proc.stdout.decode("utf-8")
				#sourceArchiveUrl: gs://functions-163510-bucket/function-1-2017-04-03T10:10:25.959Z.zip
				for line in out.split("\n"):
					key, *value = line.split(":")
					value = ":".join(value)
					if key == "sourceArchiveUrl":
						funccode = value.strip()
						codezip = os.path.join(snafulib.snafu.SnafuImport.functiondir, "..", os.path.basename(funccode))
						proc = subprocess.run("gsutil -q cp {} {}".format(funccode, codezip), stdout=subprocess.PIPE, shell=True)

						# FIXME: read and assign runtime
						env = "python"

						if target == "snafu":
							self.export_snafu_zip(codezip, None, funcname, env)
						elif target == "funktion":
							self.export_funktion(codezip, None, funcname, env)
						elif target == "fission":
							self.export_fission(codezip, None, funcname, env)
						elif target == "kubeless":
							self.export_kubeless(codezip, None, funcname, env)

	def import_lambda(self, target, convert):
		proc = subprocess.run("aws lambda list-functions", stdout=subprocess.PIPE, shell=True)
		out = proc.stdout.decode("utf-8")
		functions = json.loads(out)
		functiondir = snafulib.snafu.SnafuImport.functiondir
		for func in functions["Functions"]:
			funcname = func["FunctionName"]
			print("import", funcname)
			snafulib.snafu.SnafuImport.functiondir = os.path.join(functiondir, funcname)
			proc = subprocess.run("aws lambda get-function --function-name {}".format(funcname), stdout=subprocess.PIPE, shell=True)
			out = proc.stdout.decode("utf-8")
			function = json.loads(out)
			funccode = function["Code"]["Location"]
			codezip = os.path.join(snafulib.snafu.SnafuImport.functiondir, "..", funcname + ".zip")
			os.makedirs(snafulib.snafu.SnafuImport.functiondir, exist_ok=True)
			subprocess.run("wget -q -O {} '{}'".format(codezip, funccode), shell=True)

			env = None
			runtime = func["Runtime"]
			if runtime == "python2.7":
				env = "python"
			else:
				print("- skip ({})".format(runtime))

			if env:
				if target == "snafu":
					self.export_snafu_zip(codezip, None, funcname, env, func, convert)
				elif target == "funktion":
					self.export_funktion(codezip, None, funcname, env)
				elif target == "fission":
					self.export_fission(codezip, None, funcname, env)
				elif target == "kubeless":
					self.export_kubeless(codezip, None, funcname, env)

class SnafuImportRunner:
	def __init__(self):
		siu = SnafuImportUtility()
		siu.start_import()



# python3 setup.py sdist
# python3 setup.py bdist_wheel
# python3 setup.py bdist_egg
# twine upload dist/*.* [-r testpypi]
# -> automated in tools/make-dist.sh

from setuptools import setup
import os

def findfiles(pathlist):
	for path in pathlist:
		for dirpath, dirnames, files in os.walk(path):
			for f in files:
				if not f.endswith(".pyc") and not f.endswith(".class"):
					yield os.path.join(dirpath, f)

#import sys
#base = "lib/python{}.{}/site-packages".format(sys.version_info[0], sys.version_info[1])

setup(
	name="snafu",
	description="Swiss Army Knife of Serverless Computing",
	version="0.0.0.post3",
	url="https://github.com/serviceprototypinglab/snafu",
	author="Josef Spillner",
	author_email="josef.spillner@zhaw.ch",
	license="Apache 2.0",
	classifiers=[
		"Development Status :: 2 - Pre-Alpha",
		"Environment :: Console",
		"Environment :: No Input/Output (Daemon)",
		"Intended Audience :: Science/Research",
		"License :: OSI Approved :: Apache Software License",
		"Programming Language :: Python :: 3",
		"Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware"
	],
	keywords="cloud faas serverless functions",
	packages=["snafulib"],
	scripts=["snafu", "snafu-import", "snafu-accounts", "snafu-control"],
	#data_files=[(os.path.join(base, os.path.dirname(f)), [f]) for f in findfiles(["executors", "loggers", "authenticators", "connectors"])],
	package_data={"snafulib": findfiles(["executors", "loggers", "authenticators", "connectors"])},
	install_requires=["flask"]
	#entry_points={
	#	"console_scripts": [
	#		"snafu=snafu:main",
	#		"snafu-import=snafuimport:main",
	#		"snafu-control=snafucontrol:main",
	#		"snafu-accounts=snafuaccounts:main"
	#	]
	#}
)



# Snafu: Snake Functions - Filesystem Connector

import pyinotify
import threading
import os
import configparser

gcb = None
gf = None

class EventHandler(pyinotify.ProcessEvent):
	def process_IN_CREATE(self, event):
		event = {"file": event.pathname, "action": "create"}
		gcb(gf, event=event)

	def process_IN_DELETE(self, event):
		event = {"file": event.pathname, "action": "delete"}
		gcb(gf, event=event)

def initinternal(function, configpath):
	global gcb
	gf = function

	connecturl = None
	if not configpath:
		configpath = "snafu.ini"
	if not function:
		function = "snafu"
	if os.path.isfile(configpath):
		config = configparser.ConfigParser()
		config.read(configpath)
		if function in config and "connector.filesystem" in config[function]:
			connecturl = config[function]["connector.filesystem"]

	if connecturl:
		wm = pyinotify.WatchManager()
		handler = EventHandler()
		notifier = pyinotify.Notifier(wm, handler)
		mask = pyinotify.IN_DELETE | pyinotify.IN_CREATE
		wdd = wm.add_watch(connecturl, mask, rec=True)
		notifier.loop()

def init(cb, function=None, configpath=None):
	global gcb
	gcb = cb

	t = threading.Thread(target=initinternal, daemon=True, args=(function, configpath))
	t.start()
